# Copyright (C) 2021-2024 by Vd.
# This file is part of Snowflake package.
# Snowflake is released under the MIT License (see LICENSE).


from time import time

import pytest

from snowflake import Snowflake, SnowflakeGenerator
from snowflake import MAX_SEQ, MAX_INSTANCE


def test_generator():
    gen = SnowflakeGenerator.from_snowflake(Snowflake.parse(856165981072306191, 1288834974657))

    assert gen.epoch == 1288834974657

    val = next(gen)

    sf = Snowflake.parse(val, 1288834974657)

    assert sf.instance == 363
    assert sf.epoch == 1288834974657
    assert sf.seq == 0

    assert str(sf.timedelta) == "14917 days, 1:42:54.657000"

    assert sf.value == val

    assert int(sf) == val


def test_gen_many():
    gen = SnowflakeGenerator(MAX_INSTANCE - 1, epoch=4096)

    assert gen.epoch == 4096

    s = set()

    for _, val in zip(range(MAX_SEQ), gen):
        assert Snowflake.parse(val).timestamp > 0
        assert Snowflake.parse(val).instance == MAX_INSTANCE - 1
        s.add(val)

    assert len(s) == MAX_SEQ


def test_instance_overflow():
    SnowflakeGenerator(0)

    SnowflakeGenerator(MAX_INSTANCE)

    with pytest.raises(ValueError, match="instance must not be negative and must be less than 1023!"):
        SnowflakeGenerator(-1)

    with pytest.raises(ValueError, match="instance must not be negative and must be less than 1023!"):
        SnowflakeGenerator(MAX_INSTANCE + 1)


def test_seq_overflow():
    SnowflakeGenerator(0, seq=0)

    SnowflakeGenerator(0, seq=MAX_SEQ)

    with pytest.raises(ValueError, match="seq must not be negative and must be less than 4095!"):
        SnowflakeGenerator(0, seq=-1)

    with pytest.raises(ValueError, match="seq must not be negative and must be less than 4095!"):
        SnowflakeGenerator(0, seq=MAX_SEQ + 1)


def test_epoch_overflow():
    SnowflakeGenerator(0, epoch=0)

    SnowflakeGenerator(0, epoch=int(time() * 1000))

    with pytest.raises(ValueError, match=r"epoch must not be negative and must be lower than current time (\d+)!"):
        SnowflakeGenerator(0, epoch=-1)

    with pytest.raises(ValueError, match=r"epoch must not be negative and must be lower than current time (\d+)!"):
        SnowflakeGenerator(0, epoch=int(time() * 1000 + 10000))


def test_timestamp_overflow():
    SnowflakeGenerator(0, timestamp=0)

    SnowflakeGenerator(0, timestamp=int(time() * 1000))

    with pytest.raises(ValueError, match=r"timestamp must not be negative and must be less than (\d+)!"):
        SnowflakeGenerator(0, timestamp=-1)

    with pytest.raises(ValueError, match=r"timestamp must not be negative and must be less than (\d+)!"):
        SnowflakeGenerator(0, timestamp=int(time() * 1000 + 10000))



# Copyright (C) 2021-2024 by Vd.
# This file is part of Snowflake package.
# Snowflake is released under the MIT License (see LICENSE).


from dataclasses import dataclass
from time import time
from typing import Optional
from datetime import datetime, timedelta, tzinfo

__all__ = ('Snowflake', 'SnowflakeGenerator', 'MAX_TS', 'MAX_INSTANCE', 'MAX_SEQ')

MAX_TS = 0b11111111111111111111111111111111111111111
MAX_INSTANCE = 0b1111111111
MAX_SEQ = 0b111111111111


@dataclass(frozen=True)
class Snowflake:
    timestamp: int
    instance: int
    epoch: int = 0
    seq: int = 0

    def __post_init__(self):
        if self.epoch < 0:
            raise ValueError(f"epoch must be greater than 0!")

        if self.timestamp < 0 or self.timestamp > MAX_TS:
            raise ValueError(f"timestamp must not be negative and must be less than {MAX_TS}!")

        if self.instance < 0 or self.instance > MAX_INSTANCE:
            raise ValueError(f"instance must not be negative and must be less than {MAX_INSTANCE}!")

        if self.seq < 0 or self.seq > MAX_SEQ:
            raise ValueError(f"seq must not be negative and must be less than {MAX_SEQ}!")

    @classmethod
    def parse(cls, snowflake: int, epoch: int = 0) -> 'Snowflake':
        return cls(
            epoch=epoch,
            timestamp=snowflake >> 22,
            instance=snowflake >> 12 & MAX_INSTANCE,
            seq=snowflake & MAX_SEQ
        )

    @property
    def milliseconds(self) -> int:
        return self.timestamp + self.epoch

    @property
    def seconds(self) -> float:
        return self.milliseconds / 1000

    @property
    def datetime(self) -> datetime:
        return datetime.utcfromtimestamp(self.seconds)

    def datetime_tz(self, tz: Optional[tzinfo] = None) -> datetime:
        return datetime.fromtimestamp(self.seconds, tz=tz)

    @property
    def timedelta(self) -> timedelta:
        return timedelta(milliseconds=self.epoch)

    @property
    def value(self) -> int:
        return self.timestamp << 22 | self.instance << 12 | self.seq

    def __int__(self) -> int:
        return self.value


class SnowflakeGenerator:
    def __init__(self, instance: int, *, seq: int = 0, epoch: int = 0, timestamp: Optional[int] = None):

        current = (time() * 1000.).__int__()

        if current - epoch >= MAX_TS:
            raise OverflowError(f"The maximum current timestamp has been reached in selected epoch,"
                                f"so Snowflake cannot generate more IDs!")

        timestamp = timestamp or current

        if timestamp < 0 or timestamp > current:
            raise ValueError(f"timestamp must not be negative and must be less than {current}!")

        if epoch < 0 or epoch > current:
            raise ValueError(f"epoch must not be negative and must be lower than current time {current}!")

        self._epo = epoch
        self._ts = timestamp - self._epo

        if instance < 0 or instance > MAX_INSTANCE:
            raise ValueError(f"instance must not be negative and must be less than {MAX_INSTANCE}!")

        if seq < 0 or seq > MAX_SEQ:
            raise ValueError(f"seq must not be negative and must be less than {MAX_SEQ}!")

        self._inf = instance << 12
        self._seq = seq

    @classmethod
    def from_snowflake(cls, sf: Snowflake) -> 'SnowflakeGenerator':
        return cls(sf.instance, seq=sf.seq, epoch=sf.epoch, timestamp=sf.timestamp)

    @property
    def epoch(self) -> int:
        return self._epo

    def __iter__(self):
        return self

    def __next__(self) -> Optional[int]:
        current = (time() * 1000.).__int__() - self._epo

        if current >= MAX_TS:
            raise OverflowError(f"The maximum current timestamp has been reached in selected epoch,"
                                f"so Snowflake cannot generate more IDs!")

        if self._ts == current:
            if self._seq == MAX_SEQ:
                return None
            self._seq += 1
        elif self._ts > current:
            return None
        else:
            self._seq = 0

        self._ts = current

        return self._ts << 22 | self._inf | self._seq



# Copyright (C) 2021-2024 by Vd.
# This file is part of Snowflake package.
# Snowflake is released under the MIT License (see LICENSE).


from datetime import tzinfo, timedelta

import pytest

from snowflake import Snowflake
from snowflake import MAX_TS, MAX_SEQ, MAX_INSTANCE


def test_parse():
    class UTC0531(tzinfo):
        _offset = timedelta(seconds=19860)
        _dst = timedelta(0)

        def utcoffset(self, dt):
            return self.__class__._offset

        def dst(self, dt):
            return self.__class__._dst

    sf = Snowflake.parse(856165981072306191, 1288834974657)

    assert sf.timestamp == 204125876682
    assert sf.instance == 363
    assert sf.epoch == 1288834974657
    assert sf.seq == 15

    assert sf.seconds == 1492960851.339
    assert sf.milliseconds == 1492960851339

    assert str(sf.datetime) == "2017-04-23 15:20:51.339000"
    assert str(sf.datetime_tz(UTC0531())) == "2017-04-23 20:51:51.339000+05:31"

    assert str(sf.timedelta) == "14917 days, 1:42:54.657000"

    assert sf.value == 856165981072306191

    assert int(sf) == 856165981072306191


def test_min():
    assert Snowflake(0, 0).value == 0


def test_max():
    assert Snowflake(MAX_TS, MAX_INSTANCE, seq=MAX_SEQ).value == 9223372036854775807


def test_timestamp_overflow():
    Snowflake(0, 0)

    Snowflake(MAX_TS, 0)

    with pytest.raises(ValueError, match="timestamp must not be negative and must be less than 2199023255551!"):
        Snowflake(MAX_TS + 1, 0)

    with pytest.raises(ValueError, match="timestamp must not be negative and must be less than 2199023255551!"):
        Snowflake(-1, 0)


def test_instance_overflow():
    Snowflake(0, 0)

    Snowflake(0, MAX_INSTANCE)

    with pytest.raises(ValueError, match="instance must not be negative and must be less than 1023!"):
        Snowflake(0, -1)

    with pytest.raises(ValueError, match="instance must not be negative and must be less than 1023!"):
        Snowflake(0, MAX_INSTANCE + 1)


def test_epoch_overflow():
    Snowflake(0, 0, epoch=0)

    with pytest.raises(ValueError, match="epoch must be greater than 0!"):
        Snowflake(0, 0, epoch=-1)


def test_seq_overflow():
    Snowflake(0, 0, seq=0)

    Snowflake(0, 0, seq=MAX_SEQ)

    with pytest.raises(ValueError, match="seq must not be negative and must be less than 4095!"):
        Snowflake(0, 0, seq=-1)

    with pytest.raises(ValueError, match="seq must not be negative and must be less than 4095!"):
        Snowflake(0, 0, seq=MAX_SEQ + 1)



# Copyright (C) 2021-2024 by Vd.
# This file is part of Snowflake package.
# Snowflake is released under the MIT License (see LICENSE).


from .snowflake import Snowflake, SnowflakeGenerator
from .snowflake import MAX_TS, MAX_SEQ, MAX_INSTANCE

VERSION = "v1.0.2"
__version__ = VERSION


def version():
    return VERSION
